#include "moglclass.h"
#include "shaders.h"
#include "geometryPositionNormalTexture.h"
#include "light.h"
#include "material.h"
#include "texture2D.h"
#include "camera.h"
#include "arcball.h"

using namespace moglclass;

#define NUMSHADERS 4

Shader shShaders[NUMSHADERS];
Program spMain;

float g_fovy = 45.0f;
float g_aspect = 1.0f;
float g_zNear = 1.0f;
float g_zFar = 100.f;

Camera camera1(FLY);
//float cameraHeading = 0.0f;
//float cameraZ = 10.0f;
glm::vec3 g_eye(0, 0, 10);
glm::vec3 g_at(0, 0, 0);
glm::vec3 g_up(0, 1, 0);

glm::mat4 Identity(1.0f);
glm::mat4 World(1.0f);
glm::mat4 View(1.0f);
glm::mat4 Projection(1.0f);
glm::mat4 MVP(1.0f);
glm::mat4 xyRotation(1.0f);

Light light0(glm::vec4(4.0f, 4.0f, 4.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f)); // directional light
Light light1(glm::vec4(-5.0f, 5.0f, 4.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f)); // point light
Light light2(glm::vec4(0.0f, 0.0f, -1.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 0.0f, 1.0f), 0.0f, 1.0f, 0.0f, 60.0f, 50.0f, glm::vec3(0.0f, 0.0f, -1.0f)); // spot light

Material whiteMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);
Material redMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 0.0f, 0.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);
Material greenMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(0.0f, 1.0f, 0.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);

Texture2D * openglTex = NULL;
Texture2D * mainTex = NULL;

std::vector<GeometryPositionNormalTexture*> geo;
GeometryPositionNormalTexture * sph = NULL;
GeometryPositionNormalTexture * cyl = NULL;
int g_geomode = 0;
bool g_wiremode = false;

Arcball arcball1(SCREEN_WIDTH, SCREEN_HEIGHT);
bool g_mousemove = false;

void init( void )
{
    // load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("MTexturePerPixelShading.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("MTexturePerPixelShading.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	Projection = glm::perspective(g_fovy, g_aspect, g_zNear, g_zFar);
	View = camera1.lookAt(g_eye, g_at, g_up);

	openglTex = new Texture2D("opengl.jpg");
	mainTex = new Texture2D("brick.jpg");

	geo.push_back(new Cone(glm::vec3(0, -1.5, 1), 2.0f, 4.0f, 4));
	geo.push_back(new Cube(glm::vec3(0, 0, 0), 3.0f));
	geo.push_back(new Sphere(glm::vec3(-2, 0, 0), 1.5f, 16, 16));
	geo.push_back(new Cylinder(glm::vec3(-1, 0, 0), 2.0f, 4.0f, 16));
	geo.push_back(new Torus(glm::vec3(0, 0, 0), 2.0f, 0.6f, 32, 16));
	geo.push_back(new Parallelepiped(glm::vec3(-2.0f, -1.0f, 0.0f), glm::vec3(1.2f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.2f), glm::vec3(2.5f, 3.0f, 0.0f)));
	geo.push_back(new Grid(glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f), glm::vec3(0.0f, 1.0f, 0.0f), 8.0f, 6.0f, 4, 3));

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void cleanup()
{
	if (openglTex) delete openglTex;
	if (mainTex) delete mainTex;

	spMain.deleteProgram();
	printf("clean up\n");
}

void drawGeometry(glm::mat4& worldMatrix)
{
	//World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", worldMatrix);
	spMain.setUniform("gTextureSampler", 0);
	mainTex->bind();
	geo[g_geomode]->draw();
	mainTex->disable();
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );     // clear the window

	// set drawmode
	if (g_wiremode) 
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	// MVP matrix
	View = camera1.View();
	//World = glm::rotate(glm::mat4(1.0f), -cameraHeading, glm::vec3(0, 1, 0)) * glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, -cameraZ));
	World = arcball1.getRotationMatrix() * xyRotation;

	MVP = Projection * View * World;
	spMain.useProgram();
	spMain.setUniform("gMVP", MVP);
	spMain.setUniform("gProjection", Projection);
	spMain.setUniform("gView", View);
	spMain.setUniform("gModel", World);

	glm::vec3 cameraPos = camera1.getPosition();
	//printf("cameraPos: %f %f %f\n", cameraPos[0], cameraPos[1], cameraPos[2]);
	spMain.setUniform("gEyePosition", cameraPos);

	// light
	light0.setUniform(&spMain, "gLight0");
	light1.setUniform(&spMain, "gLight1");
	light2.setUniform(&spMain, "gLight2");
	// material
	whiteMaterial.setUniform(&spMain, "gMaterial");
	// geometry
	drawGeometry(World);

    glutSwapBuffers();
}

void update()
{
	static int currentTime, deltaTime, prevTime = 0; 
	static int startTime, elapsedTime = 0;
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	deltaTime = currentTime - prevTime;
	prevTime = currentTime;

	// translate & rotate around x/y axis
	static float angle = 0.0f;
	angle += 0.001f;

	// geometry: translate & rotate in x & rotate in y
	//glm::mat4 T = glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, 1));
	//glm::mat4 R = glm::rotate(T, angle, glm::vec3(1, 0, 0));
	//xyRotation = glm::rotate(R, angle, glm::vec3(0, 1, 0));
	xyRotation = glm::rotate(glm::mat4(1.0f), angle, glm::vec3(0, 1, 0));

	elapsedTime = currentTime - startTime;
	glutPostRedisplay();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case ' ':
		g_wiremode = !g_wiremode;
		break;  
	case 'g':
		g_geomode = (g_geomode + 1) % (int)geo.size();
		printf("g_geomode=%d\n", g_geomode);
		break;
	case 'r':
		arcball1.reset();
		break;
	}
	glutPostRedisplay();
}
/*
void specialkey(int key, int x, int y )
{
	switch (key)  
	{    
	case GLUT_KEY_LEFT:
		cameraHeading -= 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_RIGHT:    
		cameraHeading += 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_UP:      
		cameraZ += 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_DOWN:
		cameraZ -= 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_PAGE_UP:      
		break;  
	case GLUT_KEY_PAGE_DOWN:
		break;  
	}
	glutPostRedisplay();
}
*/

void specialkey(int key, int x, int y)
{
 if (key == GLUT_KEY_F1)   // x-movement
        camera1.strafe(0.5);
 else if (key == GLUT_KEY_F2)
        camera1.strafe(-0.5);
 else if (key == GLUT_KEY_F3) // y-movement
        camera1.fly(0.5);
 else if (key == GLUT_KEY_F4)
        camera1.fly(-0.5);
 else if (key == GLUT_KEY_F5) // z-movement
        camera1.walk(0.5);
    else if (key == GLUT_KEY_F6)
        camera1.walk(-0.5);

 else if (key == GLUT_KEY_F7) // yaw (by y-axis)
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_F8)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_F9) // pitch (by x-axis)
        camera1.pitch(2.5);
    else if (key == GLUT_KEY_F10)
        camera1.pitch(-2.5);
 else if (key == GLUT_KEY_HOME) // roll (by z-axis)
        camera1.roll(2.5);
    else if (key == GLUT_KEY_END)
        camera1.roll(-2.5);

 else if (key == GLUT_KEY_LEFT) // same as town
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_RIGHT)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_UP) 
        camera1.walk(0.5);
    else if (key == GLUT_KEY_DOWN)
        camera1.walk(-0.5);

 glutPostRedisplay();
}


void mouse(int button, int state, int mx, int my)
{
	if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
		exit(0);
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
	{
		g_mousemove = true;
		arcball1.startRotation(mx, my);
	}
	else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
	{
		g_mousemove = false;
		arcball1.stopRotation();
	}
}

void motion(int mx, int my)
{
	int w = glutGet(GLUT_WINDOW_WIDTH);
	int h = glutGet(GLUT_WINDOW_HEIGHT);
	//float x = (float) (mx - w*0.5) / w;
	//float y = (float) (h*0.5 - my) / h;
	//fprintf(stderr, "x=%f y=%f\n", x, y);

	if (g_mousemove)
	{
		arcball1.updateRotation(mx, my);
	}

	glutPostRedisplay();
}

void reshape(int w, int h)
{
	g_aspect = (float) (w/h);

    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glutPostRedisplay();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutInitWindowSize( SCREEN_WIDTH, SCREEN_HEIGHT );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
    glutSpecialFunc( specialkey );
	glutMouseFunc( mouse ); 
	glutMotionFunc( motion ); 
	glutIdleFunc( update );
	glutReshapeFunc( reshape );

    glutMainLoop();

	cleanup();

    return 0;
}
